(** Time-stamp: <modified the 02/04/2013 (at 08:32) by Erwan Jahier> *)
 
open Lxm
open Lic

let dbg = Some(Verbose.get_flag "exec")

type action = ActionsDeps.action

(* Raised when a soc that haven't been translated yet is used in
   another soc during the translation *)
exception Undef_soc of Soc.key * Lxm.t * Lic.by_pos_op * Soc.var_type list

(*********************************************************************************)
(** Informations lies au contexte de traduction. *)
type ctx    = {
  prg           : LicPrg.t;
  last_temp_var : int;
  last_mem      : int;
  locals        : Soc.var list;
}

let create_context: (LicPrg.t -> ctx) =
  fun prg ->
    {
      prg = prg;
      last_temp_var = 0;
      last_mem      = 0;
      locals        = [];
    }

let rec lic_to_soc_type: (Lic.type_ -> Soc.var_type) = 
  function
  | Lic.Bool_type_eff -> Soc.Bool
  | Lic.Int_type_eff  -> Soc.Int
  | Lic.Real_type_eff -> Soc.Real
  | Lic.External_type_eff s -> Soc.Extern (Ident.string_of_long s)
  | Lic.Enum_type_eff     (id, l) -> (
    Soc.Enum(Ident.string_of_long id, List.map Ident.string_of_long l)
  )
  | Lic.Struct_type_eff   (id, fl) -> (
    let trans_field (id,(t,_)) = (* fde_value is ignored. Good idea? *)
      Ident.to_string id, lic_to_soc_type t
    in
    let id = Ident.string_of_long id in
    Soc.Struct(id, List.map trans_field fl)
  )
  | Lic.Array_type_eff(ty,i) -> Soc.Array(lic_to_soc_type ty,i)
  | Lic.Abstract_type_eff (id, _) -> assert false
  | Lic.TypeVar Lic.Any -> Soc.Alpha 0
  | Lic.TypeVar Lic.AnyNum -> Soc.Alpha 1



(*********************************************************************************)
(** Renomme une variable dfinie par l'utilisateur.

    On veut viter de crer des variables temporaires portant le mme nom que
    celles dfinies par l'utilisateur. Donc on renomme simplement celles de
    l'utilisateur, c'est le plus simple. 

XXX obselete ?
Mieux vaudrait utiliser le meme mechanisme que celui utilis
actuellement lors des l2l*.ml 
???
*)

 (* R1: je vire le "_" parce que ca m'emmerde ce truc. 
                        On verra plus tard les clash de nom...*)
let user_var_prefix = "_"
let user_var_prefix = ""

let rename_user_var: (string -> string) = fun s ->
  let prefix = user_var_prefix in
  let suffix = "" in
  prefix ^ s ^ suffix

let is_predefined_const: string -> Lic.type_ option = 
  function
    | "true" | "false" -> Some Lic.Bool_type_eff
    | _ -> None

(*********************************************************************************)
(* Returns the list of indexes represented by the slice *)
let (slice_info_to_index_list : Lic.slice_info -> int list) =
  fun si -> 
    let (f,l,s) = (si.Lic.se_first, si.Lic.se_last, si.Lic.se_step) in
    let rec aux f =
      if f>l && s > 0 || f<l && s <0 then [] else
        f::(aux (f+s))
    in
      aux f

(* Returns Some(thing) if val_exp is a leaf (a var or a constant) 

XXX c'est pas tres clair le role de cette fonction. Expliquer !
*)
let rec get_leaf: (LicPrg.t -> Lic.val_exp -> Soc.var_expr list option) = 
  fun licprg val_exp  ->
    let v = val_exp.Lic.ve_core in
    let type_ = val_exp.Lic.ve_typ in
    match v with
      | Lic.CallByNameLic(by_name_op_flg,fl) -> None
      | Lic.Merge(c_flg, cl) -> None
      | Lic.CallByPosLic (by_pos_op_flg, val_exp_list) -> (
        match by_pos_op_flg.it with
          | Lic.VAR_REF name -> 
            let type_ = (List.hd type_) in
            let translation =
              match is_predefined_const name with
                | Some type_ -> Soc.Const(name,  lic_to_soc_type type_)
                | None -> Soc.Var(rename_user_var name, lic_to_soc_type type_)
            in
            Some [translation]
          | Lic.CONST_REF l -> (
            let type_ = lic_to_soc_type (List.hd type_) in
            Some [Soc.Const(Ident.string_of_long l, type_)]
          )
          | Lic.CONST c -> assert false

(*           | Lic.PREDEF_CALL AstPredef.TRUE_n      -> Some [Soc.Const("true", Soc.Bool)] *)
(*           | Lic.PREDEF_CALL AstPredef.FALSE_n     -> Some [Soc.Const("false", Soc.Bool)] *)
(*           | Lic.PREDEF_CALL AstPredef.RCONST_n id -> Some [Soc.Const(id, Soc.Real)] *)
(*           | Lic.PREDEF_CALL AstPredef.ICONST_n id -> Some [Soc.Const(id, Soc.Int)] *)

          | Lic.STRUCT_ACCESS(field) -> (
            let expr = match val_exp_list with [e] -> e | _ -> assert false in
            let type_ = lic_to_soc_type (List.hd type_) in
            let filter_expr = match get_leaf licprg expr with
              | Some [f] -> f
              | None -> assert false
              | _ ->  assert false
            in
            Some [Soc.Field(filter_expr, field, type_)]
          )
          | Lic.ARRAY_ACCES i -> (
            let expr = match val_exp_list with [e] -> e | _ -> assert false in
            let type_ = lic_to_soc_type (List.hd type_) in
            let filter_expr = match get_leaf licprg expr with
              | Some [f] -> f
              | None -> assert false  (* should not happen, since the expression should be a leaf *)
              | _ -> assert false (* We should get only ONE filter, otherwise it doesn't make any
                                     sense *)
            in
            Some [Soc.Index(filter_expr, i, type_)]
          )
          | Lic.TUPLE -> (
            let var_values = List.map (get_leaf licprg) val_exp_list in
            let del_some = function | None -> assert false | Some x -> x in
            Some (List.flatten (List.map del_some var_values))
          )
          | Lic.ARRAY_SLICE si -> (
            let id = match val_exp_list with 
              | [{Lic.ve_core=Lic.CallByPosLic({it=Lic.VAR_REF id},[])}] -> id
              | _ -> assert false
            in
            let type_elt_ref,type_ref = 
              match type_ with
                | [Lic.Array_type_eff(t,i)]  -> 
                  let t_soc = lic_to_soc_type t in
                  t_soc,  Soc.Array(t_soc,i)
                | _ -> assert false (* should not occur *)
            in
            let index_list = slice_info_to_index_list si in
            let exploded_array =  
              (* val_exp is a var ident (t) of type array; we want to gen the list
                 t[i1], ...,t[in], where the index are specified by the slice 
              *)
              List.map
                (fun i -> Soc.Index(Soc.Const(id, type_ref), i, type_elt_ref))
                index_list
            in
            Some(exploded_array)
          )
          | Lic.PREDEF_CALL _
          | Lic.CALL _
          | Lic.PRE
          | Lic.ARRAY 
          | Lic.HAT _
          | Lic.ARROW
          | Lic.FBY
          | Lic.CURRENT
          | Lic.WHEN(_)
          | Lic.CONCAT
            -> None
      )      
(** Traduction d'une partie gauche d'quation en filtre d'accs soc. *)
let rec filter_of_left_part: (LicPrg.t -> Lic.left -> Soc.var_expr list) = 
  fun licprg lp -> 
    let type_ = Lic.type_of_left lp in
    match lp with
      | Lic.LeftVarLic  (vi, _lxm) -> (
        [Soc.Var (rename_user_var vi.Lic.var_name_eff, lic_to_soc_type vi.Lic.var_type_eff)]
      )
      | Lic.LeftFieldLic(lp,field,t) -> (
        let lpl = filter_of_left_part licprg lp in
        List.map (fun lp -> Soc.Field(lp, field, lic_to_soc_type t)) lpl
      )
      | Lic.LeftArrayLic(lp,index,t) -> (
        let lpl = filter_of_left_part licprg lp in
        List.map (fun lp -> Soc.Index(lp, index, lic_to_soc_type t (* type_ ? *))) lpl
      )
      | Lic.LeftSliceLic(lp,si,t) -> (
      (* we expand left part slices *)
        let lpl = filter_of_left_part licprg lp in
        let index_list = slice_info_to_index_list si in
        List.flatten (List.map (
          fun lp -> List.map (fun index -> Soc.Index(lp, index, lic_to_soc_type t)) index_list) lpl)
      )

(*********************************************************************************)
let rec (gao_of_action: action -> Soc.gao) = 
  fun (ck, il, ol, op, lxm) ->
    let rec unpack_clock = function
      | Lic.BaseLic -> Soc.Call (ol, op, il)
      | Lic.ClockVar i -> (* should not occur ?*)
        Soc.Call (ol, op, il)
      | Lic.On((c, value), inner_clock) ->
(*         let inner_clock = match inner_clock_opt with *)
(*           | Some x -> x *)
(*           | None ->  *)
(*                 (* TODO? Retreive the clock of c *) *)
(*             Errors.internal lxm; *)
(*             assert false  *)
(*         in *)
        Soc.Case (Ident.string_of_idref c, [value, [unpack_clock inner_clock]] )
    in
    unpack_clock ck

let build_step: Lxm.t -> string -> Lic.node_exp -> Soc.var list -> 
  action list -> Soc.step_method =
  fun lxm name node locals actions ->
    (* Converti les entres/sorties d'un noeud en index
       d'entres/sorties du composant *)
    let convert_node_interface = fun l ->
      fst (List.fold_left (fun (a, i) _ -> a @ [i], i+1) ([], 0) l)
    in
      {
        Soc.name    = name;
        Soc.lxm     = lxm;
        Soc.idx_ins  = convert_node_interface node.Lic.inlist_eff;
        Soc.idx_outs = convert_node_interface node.Lic.outlist_eff; 
        Soc.impl     = Soc.Gaol (locals, List.map gao_of_action actions)
      }

let (lic_to_soc_var : Lic.var_info -> Soc.var) = 
  fun vi -> 
    vi.Lic.var_name_eff, lic_to_soc_type vi.Lic.var_type_eff

let soc_profile_of_node: Lic.node_exp -> Soc.var list * Soc.var list =
  fun n ->
    let inputs  = List.map lic_to_soc_var n.Lic.inlist_eff in
    let outputs = List.map lic_to_soc_var n.Lic.outlist_eff in
    inputs, outputs 

let (make_soc_key_of_node_exp : Lic.node_key -> Soc.var_type list -> Soc.key) =
fun nk vl -> 
  LicDump.string_of_node_key_rec nk, vl, None

let (soc_key_of_node_exp : Lic.node_exp -> Soc.key) =
  fun n -> 
    let svi,svo = soc_profile_of_node n in
    let sk = make_soc_key_of_node_exp n.node_key_eff (List.map snd  (svi@svo)) in
    sk

(* XXX duplicated code with get_leaf *)
let (val_exp_to_filter: LicPrg.t -> Lic.val_exp -> Soc.var_expr) =
  fun licprg val_exp ->
    let v = val_exp.Lic.ve_core in
    let type_ = val_exp.Lic.ve_typ in
    match v with
      | CallByNameLic(by_name_op_flg,fl) -> assert false (* should not occur *)
      | Merge(c_flg, cl) -> 
        print_string "Merge not yet supported, sorry\n";
        flush stdout;
        assert false
      | CallByPosLic (by_pos_op_flg, val_exp_list) -> (
        match by_pos_op_flg.it with
          | VAR_REF name -> 
            let type_ = (List.hd type_) in
            let translation =
              match is_predefined_const name with
                | Some type_ -> Soc.Const(name,  lic_to_soc_type type_)
                | None -> Soc.Var(rename_user_var name, lic_to_soc_type type_)
            in
            translation
          | CONST_REF l -> (
            let type_ = lic_to_soc_type (List.hd type_) in
            Soc.Const(Ident.string_of_long l, type_)
          )
          | CONST (Bool_const_eff true) -> Soc.Const("true", Soc.Bool)
          | CONST (Bool_const_eff false) -> Soc.Const("false", Soc.Bool)
          | CONST (Int_const_eff i)  -> Soc.Const(i, Soc.Int)
          | CONST (Real_const_eff str) -> Soc.Const(str, Soc.Real)
          | CONST Extern_const_eff  _ -> assert false
          | CONST Abstract_const_eff  _ -> assert false
          | CONST Enum_const_eff   _ -> assert false
          | CONST  Struct_const_eff _ -> assert false
          | CONST Array_const_eff  _ -> assert false
          | CONST Tuple_const_eff _ -> assert false

          | STRUCT_ACCESS(field) -> (
            let expr = match val_exp_list with [e] -> e | _ -> assert false in
            let type_ = match type_ with [t] -> lic_to_soc_type t | _ -> assert false in
            let filter_expr = match get_leaf licprg expr with
              | Some [f] -> f
              | None -> assert false
              | _ ->  assert false
            in
            Soc.Field(filter_expr, field, type_)
          )
          | ARRAY_ACCES i -> (
            let expr = match val_exp_list with [e] -> e | _ -> assert false in
            let type_ = lic_to_soc_type (List.hd type_) in
            let filter_expr = match get_leaf licprg expr with
              | Some [f] -> f
              | None -> assert false
              | _ -> assert false
            in
            Soc.Index(filter_expr, i, type_)
          )
          | PREDEF_CALL _
          | CALL _
          | PRE
          | ARROW
          | FBY
          | CURRENT
          | WHEN(_)
          | TUPLE
          | CONCAT
          | HAT _
          | ARRAY
          | ARRAY_SLICE _ -> 
            let lxm = by_pos_op_flg.src in
            let msg = (Lxm.details lxm) ^ 
              ": only one operator per equation is allowed.\n"
            in
            raise (Errors.Global_error msg)
      )

(*********************************************************************************)
(* type instance_init = Soc.instance * action list (* instance + son initialisation *) *)

(** Cr une opration  partir d'un nom de mthode d'un composant. *)
let soc_step_to_operation: 
    Soc.ident -> Soc.t -> Soc.instance option -> Soc.atomic_operation =
  fun name comp -> function
    | None -> Soc.Procedure (comp.Soc.key) 
    | Some (i) -> Soc.Method(i,name) 

let (action_of_step : Lxm.t -> Soc.t -> Lic.clock -> Soc.var_expr list -> 
      Soc.var_expr list -> Soc.instance option -> Soc.step_method -> action) =
  fun lxm c clk il ol mem step ->
    let local_nth i l = 
      if i < 0 then 
        (* to handle the particular case of arrow. XXX just a crutch to make it works *)
        let mem_name = SocPredef.get_mem_name c.Soc.key Soc.Bool in
        Soc.Var(mem_name, Soc.Bool)
      else
      try List.nth l i 
      with _ -> 
        print_string (
          "\n*** Cannot get the " ^ (string_of_int (i+1)) ^ 
            "th element of a list of size " ^ (string_of_int (List.length l))^"\n");
        flush stdout;
        assert false 
    in
    let inputs  = List.map (fun i -> local_nth i il) step.Soc.idx_ins  in
    let outputs = List.map (fun i -> local_nth i ol) step.Soc.idx_outs in
    let call_action = soc_step_to_operation step.Soc.name c mem in
      (clk, inputs, outputs, call_action, step.Soc.lxm)

(** Cr un nouveau nom pour une instance. *)
let create_new_instance_name: (Soc.key -> ctx -> ctx * string) = fun (soc_name,_,_) ctx ->
  let prefix  = soc_name in
  let suffix  = ""  in
  let make id = Format.sprintf "%s%d%s" prefix id suffix in
  let new_ctx = {ctx with last_mem = ctx.last_mem + 1 } in
    new_ctx, make new_ctx.last_mem

 (** Cr une nouvelle instance pour tre utilise dans un composant.

     Pendant la traduction d'un oprateur, on s'apercoit que cet
     oprateur dispose d'une (ou plusieur) mmoire.  Il faut donc
     qu'on cr une instance de ce composant. *)
let create_instance_from_soc: (ctx -> Soc.t -> ctx * Soc.instance) = 
  fun ctx c ->
    let ctx, inst_name = create_new_instance_name c.Soc.key ctx in
      ctx, (inst_name, c.Soc.key)

(* if the soc has memories (a pre, or node with memory), do create an instance *)
let (make_instance : 
       Lxm.t -> Lic.clock -> ctx -> Soc.t -> ctx * Soc.instance option) =
  fun lxm clk ctx soc  -> 
    match soc.Soc.instances with
      | [] -> (
        match soc.Soc.have_mem with
          | None -> ctx, None
          | Some _ -> (* pre/fby*)
            let ctx, m = create_instance_from_soc ctx soc in
            ctx, Some(m)
      )
      | _  -> (* the soc has sub-soc with memory *)
        let ctx, m = create_instance_from_soc ctx soc in
        ctx, Some(m)

(*********************************************************************************)
(** Transforme une expression en action(s), et retourne la liste des variables
    cres pour stocker le rsultat du calcul de cette expression.

    Ces nouvelles variables serviront d'entres pour l'expression parente. 
*)
type e2a_acc = ctx * action list * Soc.var_expr list * Soc.instance list * ActionsDeps.t


(* Bquille en attendant mieux *)
let by_pos_op_to_soc_ident = function
  | PRE  -> "Lustre::pre"
  | ARROW -> "Lustre::arrow" 
  | FBY-> "Lustre::fby"
  | CURRENT -> "Lustre::current"
  | CONCAT-> "Lustre::concat"
  | ARRAY  -> "Lustre::array"
  | HAT _ -> "Lustre::hat"
  | PREDEF_CALL n
  | CALL n -> LicDump.string_of_node_key_rec n.it
  | _  -> assert false


let (get_exp_type : Soc.var_expr list -> Soc.var_type list) =
  List.map Soc.var_type_of_var_expr

let rec (actions_of_expression_acc: Lxm.t -> Soc.tbl ->  
         Lic.clock -> Soc.var_expr list -> e2a_acc -> Lic.val_exp -> e2a_acc) =
  fun lxm soc_tbl clk lpl acc expr ->
    let (ctx, al, iol, ml, deps) = acc in
    match get_leaf ctx.prg expr with
      | Some names -> 
        (* expr est dj une feuille (un ident ou une constante), RAF. *)
        let action = clk, names, lpl, Soc.Assign,lxm in
        (ctx, action::al, iol@names, ml, deps)
      | None -> (
        let v = expr.Lic.ve_core in
        match v with
          | CallByNameLic(by_name_op_flg,fl) -> (
            (* Pas de soc pour les structures non plus.  On se
               contente d'clater la structure en autant d'galits
               que ncessaire.  *)
            let lxm = by_name_op_flg.src in
            let filter_to_field filter field ftype =
              let ftype =  match ftype  with [x] -> x | _ -> assert false in
              let filter = match filter with [x] -> x | _ -> assert false in
              Soc.Field(filter, field, lic_to_soc_type ftype)
            in
            let actions = 
              List.map
                (fun (fn, fv) -> 
                  let ft = fv.ve_typ in
                  let fv = val_exp_to_filter ctx.prg fv in
                  (clk, [fv], [filter_to_field lpl fn.it ft], Soc.Assign, lxm)
                )
                fl
            in
            ctx, actions@al, iol, ml, deps
          )
          | Merge(c_flg, cl) -> 
            print_string "Merge not yet supported, sorry\n";
            flush stdout;
            assert false
          | CallByPosLic (by_pos_op_flg, val_exp_list) -> (
            match by_pos_op_flg.it with
              | Lic.ARRAY_SLICE _ | Lic.VAR_REF _ | Lic.CONST_REF _ | Lic.CONST _
              | Lic.ARRAY_ACCES _ | Lic.STRUCT_ACCESS _ | Lic.TUPLE
                -> assert false (* should not occur: handled via get_leaf *)
              | Lic.WHEN ck -> (assert false 
              (* XXX FINISH ME!!! *)
              (*                 (* L'oprateur when n'est pas un composant, il modifie *)
              (*                    simplement les conditions de traitement des expressions. *) *)
              (*                 let ctx, actions, inputs, mems, deps = *)
              (*                   actions_of_expression_list Soc.tbl clk lpl acc val_exp_list *)
              (*                 in *)
              (*                 let new_clock =  *)
              (*                   match ck with *)
              (*                     | AstCore.Base -> CE_base *)
              (*                     | AstCore.NamedClock {it=(cc,cv)} -> *)
              (*                       CE_clock(name, value, clk) *)
              (*                         Clocking.clock_eff_of_clock_exp ctx.prg ck  *)
              (*                 in *)
              (*                 let ctx, outputs, actions_reclocked =  *)
              (*                   match actions with *)
              (*                     | [] -> *)
              (*                       (* L'expression du when est une feuille, on cr quand *)
              (*                          mme une nouvelle action pour clocker la feuille. *) *)
              (*                       ctx, lpl, [new_clock, inputs, lpl, Soc.Assign, lxm] *)
              (*                     | _  -> *)
              (*                       ctx, inputs, *)
              (*                       (* Remplacement de l'horloge des actions de l'expression *)
              (*                          par la nouvelle horloge issue du `when`. *) *)
              (*                       List.map  *)
              (*                         (fun (_, i,o,op,lxm) -> new_clock,i,o,op,lxm)  *)
              (*                         actions *)
              (*                 in *)
              (*                 ctx, actions_reclocked, outputs, mems, deps *)
              )

              | HAT _
              | ARRAY 
              | CALL _  
              | PREDEF_CALL _ 
              | PRE | ARROW | FBY | CURRENT | CONCAT -> (
                (* retreive the soc of "expr" in soc_tbl *)
                let soc : Soc.t =
                  let id = by_pos_op_to_soc_ident by_pos_op_flg.it in 
                  let args_types : Soc.var_type list =
                    List.map lic_to_soc_type
                      (List.flatten (List.map (fun ve -> ve.ve_typ) val_exp_list))
                  in
                  let res_type = get_exp_type lpl in
                  let full_profile = args_types @ res_type in
                  let sk = make_soc_key_of_node_exp (("",id),[]) full_profile in
                  try Soc.SocMap.find sk soc_tbl 
                  with Not_found ->
                    Verbose.exe ~flag:dbg (fun () ->
                      let l = Soc.SocMap.bindings soc_tbl in
                      let kl = fst (List.split l) in
                      let klstr = List.map SocUtils.string_of_soc_key kl in                      
                      print_string ("\n********* Cannot find the soc.key " ^
                                       (SocUtils.string_of_soc_key sk) ^ "  in \n\t" ^
                                       (String.concat "\n\t" klstr)^"\n");
                      flush stdout;
                    );
                    raise (Undef_soc (sk, lxm, by_pos_op_flg.it, args_types))
                in
                (* Use that soc to build the corresponding
                   - actions
                   - instances
                   - action dependances
                *)
                let inputs = List.map (val_exp_to_filter ctx.prg) val_exp_list in
                let ctx, mem_opt = make_instance lxm clk ctx soc in
                let actions =
                  let m2act = action_of_step lxm soc clk inputs lpl mem_opt in
                  List.map m2act soc.Soc.step
                in
                let dependances : ActionsDeps.t =
                  let (prefixed_actions : (Soc.ident * action) list) = List.map2
                    (fun s a -> s.Soc.name,a) soc.Soc.step actions
                  in
                  ActionsDeps.generate_deps_from_step_policy
                    soc.Soc.precedences prefixed_actions
                in
                let ml = match mem_opt with Some m -> m::ml | None -> ml in
                (ctx, actions, lpl, ml, dependances)
              )
          )
      )

        
(** Traduction d'une liste d'expressions. *)
and (actions_of_expression_list: Lxm.t -> Soc.tbl -> Lic.clock -> Soc.var_expr list -> 
     e2a_acc -> Lic.val_exp list -> e2a_acc) =
  fun lxm soc_tbl clk lpl expr_list acc ->
    List.fold_left (actions_of_expression_acc lxm soc_tbl clk lpl) expr_list acc


let (actions_of_expression : Lxm.t -> Soc.tbl -> ctx -> Lic.clock -> Soc.var_expr list ->
      Lic.val_exp -> e2a_acc) =
  fun lxm soc_tbl ctx clk lpl expr ->
    let acc0 = (ctx, [], [], [], ActionsDeps.empty) in
      actions_of_expression_acc lxm soc_tbl clk lpl acc0 expr

(*********************************************************************************)
 (** Traduction d'une quation complte.

    On traduit d'abord l'expression de l'quation, puis on fait une galit
    entre les variables issues de la traduction de l'expression et la partie
    gauche de l'quation. *)
let (actions_of_equation: Lxm.t -> Soc.tbl -> ctx -> Lic.eq_info -> 
      ctx * action list * Soc.instance list * ActionsDeps.t) = 
  fun lxm soc_tbl ctx (left_part, right_part) ->
    let clk = right_part.ve_clk in
    let clk = match clk with clk::_ -> clk | _ -> assert false in
    let left_loc = List.map (filter_of_left_part ctx.prg) left_part in
    let left_loc = List.flatten left_loc in
    let ctx, actions, _, instances, deps =
      actions_of_expression lxm soc_tbl ctx clk left_loc right_part
    in
(*     let final_action = clk_l, inputs, left_loc, Soc.Identity in *)
(*     let deps = add deps final_action actions in *)
      ctx, actions, instances, deps
 


(*********************************************************************************)
open Soc
let rec f: (LicPrg.t -> Lic.node_key -> Soc.key * Soc.tbl) = 
  fun prog mnk ->
    let rec (process_node : Lic.node_key -> Soc.tbl -> Soc.key * Soc.tbl) =
      fun nk soc_tbl -> 
        let node = 
          match LicPrg.find_node prog nk with
            | None  -> 
              prerr_string (
                "*** "^ (LicDump.string_of_node_key_rec nk) ^" not defined.\n" ^
                  "*** Defined nodes are:"^ 
                  (String.concat ",\n"  
                     (List.map (fun (nk,_) -> "\""^LicDump.string_of_node_key_rec nk^"\"")
                        (LicPrg.list_nodes prog)))
              );
              assert false
            | Some node_exp -> node_exp
        in
        let sk = soc_key_of_node_exp node in
        let soc_tbl = 
          if SocMap.mem sk soc_tbl then soc_tbl else
            try
              (match LicPrg.find_node prog nk with
                | None  -> assert false 
                | Some node_def ->
                  (match soc_of_node prog node_def soc_tbl with
                    | Some(_,soc,soc_tbl) -> SocMap.add sk soc soc_tbl
                    | None ->
                      print_string ("Undefined soc : " ^ (string_of_node_key nk) ^ "\n");
                      flush stdout;
                      soc_tbl
                  )
              )
            with
              (* Il manque une dpendance, on essaie de la
                 traduire puis de retraduire le noeud courant. *)
              | Undef_soc (sk,lxm,Lic.CALL { it = nk2 }, types) ->
                let soc_tbl = snd (process_node nk2 soc_tbl) in
                snd (process_node nk soc_tbl)

              | Undef_soc (sk,lxm,pos_op, types) ->
                let soc = SocPredef.soc_interface_of_pos_op lxm pos_op types in
                if sk<>soc.key then (
                  print_string ("Soc key mismatch :\n\t" ^
                                   (SocUtils.string_of_soc_key sk) ^ "\n<>\n\t" ^
                                   (SocUtils.string_of_soc_key soc.key) ^ "\n");
                  flush stdout;
                  assert false
                );
                let soc_tbl = SocMap.add soc.key soc soc_tbl in
                let t = List.hd types in
                (* The arrow is translated into a if. So we make sure that the "if"
                   is in the soc tbl *)
                let if_sk =  "Lustre::if", [Bool;t;t], None in
                let soc_tbl =
                  if pos_op = Lic.ARROW && not(SocMap.mem if_sk soc_tbl) then
                    let soc = SocPredef.soc_interface_of_pos_op lxm
                      (Lic.PREDEF_CALL ({ it=("Lustre","if"),[]; src=lxm})) [Bool;t;t]
                    in
                    SocMap.add soc.key soc soc_tbl
                  else
                    soc_tbl
                in
                snd (process_node nk soc_tbl)
        in
        sk, soc_tbl
    (** Produit des soc de noeuds. *)
    and (soc_of_node: LicPrg.t -> Lic.node_exp -> Soc.tbl -> (ctx * Soc.t * Soc.tbl) option) =
      fun licprg node soc_tbl ->
        let io_list = node.Lic.inlist_eff @ node.Lic.outlist_eff in 
        let io_type = List.map (fun vi -> lic_to_soc_type vi.var_type_eff) io_list in
        let soc_key = Ident.string_of_long2 (fst node.Lic.node_key_eff), io_type, None in

        let soc_key = make_soc_key_of_node_exp node.Lic.node_key_eff io_type in
        let lxm = node.Lic.lxm in
        let ctx = create_context licprg in
        let (soc_of_body: Lic.node_body -> Soc.tbl -> (ctx * Soc.t * Soc.tbl) option) =
          fun b soc_tbl ->
            let ctx, actions, instances, deps =
              (* on itere sur la liste des quations *)
              List.fold_left
                (fun (c, a, i, d) eq ->
                  let nc, na, ni, nd = actions_of_equation eq.src soc_tbl c eq.it in
                  nc, a @ na, i @ ni, (ActionsDeps.concat nd d)
                )
                (ctx, [], [], ActionsDeps.empty)
                b.eqs_eff
            in
            (* Construction des dpendances entre les expressions *)
            let all_deps = ActionsDeps.build_data_deps_from_actions deps actions in
            Verbose.exe ~flag:dbg
              (fun () -> print_string (ActionsDeps.to_string all_deps); flush stdout);
            let actions = 
              try Toposort.f 
                    ActionsDeps.string_of_action_simple actions (ActionsDeps.find_deps all_deps)
              with Toposort.DependencyCycle(x,l) ->
                let msg = "A combinational cycle been detected "^
                  (Lxm.details lxm)^": \n  "^x^"\n  "^(String.concat "\n  " l)
                in
                raise (Errors.Global_error msg)
            in
            let (locals: Soc.var list) = 
              match node.Lic.loclist_eff with
                | None -> []
                | Some l ->  List.map (lic_to_soc_var) l 
            in
            let step = build_step lxm "step" node (locals @ ctx.locals) actions in
            let soc = {
              Soc.key         = soc_key;
              Soc.profile     = soc_profile_of_node node;
              Soc.instances   = instances ;
              Soc.step        = [step];
              Soc.have_mem = None;
              Soc.precedences = [];
            } 
            in
            Some(ctx, soc, soc_tbl)
        in
        let (soc_of_metaop: Lic.node_key -> Soc.tbl -> (ctx * Soc.t * Soc.tbl) option) =
          fun nk soc_tbl ->
            match List.sort compare (snd nk) with
              | [ConstStaticArgLic(_,Int_const_eff(c)); NodeStaticArgLic(_,iter_node)] 
              | [ConstStaticArgLic(_,Int_const_eff(c)); TypeStaticArgLic(_);
                 NodeStaticArgLic(_,iter_node)] -> (
                let nsk, soc_tbl = process_node iter_node soc_tbl in
                let nsoc = SocUtils.find lxm nsk soc_tbl in
                let nsoc_step = match nsoc.step with [s] -> s 
                  | _ -> assert false (* hmm. Iterating on a pre will not work. XXX fixme ! *)
                in
                let rec make_n_instance ctx acc n =
                  if n=0 then ctx,acc else
                    match make_instance lxm Lic.BaseLic ctx nsoc with
                      | ctx,Some inst -> make_n_instance ctx (inst::acc) (n-1)
                      | ctx,None -> ctx,[]
                in
                let c = int_of_string c in
                let ctx, instances = make_n_instance ctx [] c in 
                let soc = {
                  Soc.key       = soc_key;
                  Soc.profile   = soc_profile_of_node node;
                  Soc.instances = instances ;
                  Soc.step      = [
                    {
                      name     = "step";
                      lxm      = nsoc_step.lxm;
                      idx_ins  = nsoc_step.idx_ins;
                      idx_outs = nsoc_step.idx_outs;
                      impl     = Iterator(snd (fst nk), nsk, c);
                    }
                  ];
                  Soc.have_mem    = None;
                  Soc.precedences = [];
                } 
                in
                Some(ctx, soc, soc_tbl)
              ) 
              | _ -> 
                match (snd nk) with
                    [ConstStaticArgLic(_,Int_const_eff(i)); 
                     ConstStaticArgLic(_,Int_const_eff(j)); 
                     ConstStaticArgLic(_,Int_const_eff(k)) ] -> (
                      let i,j,k = int_of_string i, int_of_string j, int_of_string k in
                      let soc = {
                        Soc.key       = soc_key;
                        Soc.profile   = soc_profile_of_node node;
                        Soc.instances = [] ;
                        Soc.step      = [
                          {
                            name     = "step";
                            lxm      = lxm;
                            idx_ins  = [0];
                            idx_outs = [0];
                            impl     = Boolred(i,j,k);
                          }
                        ];
                        Soc.have_mem    = None;
                        Soc.precedences = [];
                      } 
                      in
                      Some(ctx, soc, soc_tbl)
                    )
                  | _ -> assert false
        in
        let (soc_of_extern: Lic.node_exp -> Soc.tbl -> (ctx * Soc.t * Soc.tbl) option) =
          fun node soc_tbl ->
            try
              let soc = SocPredef.of_soc_key soc_key in
              Some(ctx, soc, soc_tbl)
            with e -> 
              let soc = {
                Soc.key         = soc_key;
                Soc.profile     = soc_profile_of_node node;
                Soc.instances   = [] ;
                Soc.step        = [];
                Soc.have_mem    = None; (* XXX there is something todo if node.Lic.has_mem_eff; *)
                Soc.precedences = [];
              } 
              in
              (*         Some(create_context licprg, soc) *)
              print_string "Extern node not yet supported, sorry\n";
              flush stdout;
              assert false
        in
        match node.Lic.def_eff with
          | AbstractLic None -> assert false (* None if extern in the provide part *)
          | AbstractLic (Some node_exp) -> soc_of_node licprg node_exp soc_tbl
          | MetaOpLic                   -> soc_of_metaop node.Lic.node_key_eff soc_tbl
          | BodyLic b                   -> soc_of_body b soc_tbl
          | ExternLic                   -> soc_of_extern node soc_tbl

    in
    process_node mnk SocMap.empty
   


    
