(* Time-stamp: <modified the 18/03/2013 (at 17:30) by Erwan Jahier> *)



open Verbose
open AstV6
open AstCore
open Lxm
open Errors
open Parsing
open Format

let test_lex ( lexbuf ) = (
  let tk = ref (Lexer.lexer lexbuf) in 
    while !tk <> Parser.TK_EOF do
      match (Lexer.token_code !tk) with 
	  ( co , lxm ) ->
	    Printf.printf "line %3d col %2d to %2d : %15s = \"%s\"\n"
	      (line lxm) (cstart lxm) (cend lxm) co (str lxm) ;
	    tk := (Lexer.lexer lexbuf)
    done
)

(* Retourne un AstV6.t *)
let lus_load lexbuf = 
  let tree = Parser.program Lexer.lexer lexbuf in
    LicName.update_fresh_var_prefix ();
    (* ICI  *)
    AstRecognizePredef.f tree
   

type maybe_packed = 
  | Packed of AstV6.pack_or_model
  | Unpacked of AstV6.packbody 



let (get_source_list : string list -> AstV6.pack_or_model list) =
  fun infile_list -> 
    let (get_one_source : string -> string list * maybe_packed list) = 
      fun infile -> 
        let incl_files, l =
          let lexbuf = Global.lexbuf_of_file_name infile in
          if !Global.tlex then test_lex lexbuf;
          match (lus_load lexbuf) with
            | PRPackBody(incl_files, pbdy) -> incl_files, [Unpacked pbdy]
            | PRPack_or_models(incl_files, nsl) -> incl_files, (List.map (fun ns -> Packed ns) nsl)
        in
        (* If included files have a relative path, strange things may happen.
           Hence we make the path absolute, using the directory of the includer.
        *)
        let includer_dir = Filename.dirname infile in
        let fix_dir f = if Filename.is_relative f then Filename.concat includer_dir f else f in
        let incl_files = List.map fix_dir incl_files in
        incl_files, l
    in
    let rec (get_remaining_source_list : maybe_packed list * string list * string list -> 
             maybe_packed list * string list * string list) =
      fun (pack_acc, compiled, to_be_compiled) -> 
        match to_be_compiled with
          | [] -> (pack_acc, compiled, [])
          | infile::tail ->
            let infile = FilenameExtras.simplify infile in
            if List.mem infile compiled then
              get_remaining_source_list (pack_acc, compiled, tail)
            else
              let included_files, pack = get_one_source infile in
              let new_pack_acc = pack_acc@pack in
              get_remaining_source_list(
                new_pack_acc, 
                infile::compiled, 
                tail@included_files)
    in
    let infile_list = 
      (* We need absolute paths to make sure that files are not
         included several times.  Indeed, otherwise,
         FilenameExtras.simplify may miss some simplifications.  For
         example, consider the files "../../x/toto.lus" and
         "../toto.lus".  They actually refer to the same file if the
         current directory is a sub-directory of "x". Working with
         absolute paths solves the problem.
         
      *)
      let make_it_absolute f = if Filename.is_relative f then Filename.concat (Sys.getcwd ()) f else f in
      List.map make_it_absolute infile_list
    in
    let first_file = assert (infile_list <> []); List.hd infile_list in
    let included_files, first_pack = get_one_source first_file in
    let (pack_list, _compiled_files, included_files) = 
      get_remaining_source_list (first_pack, [first_file], (List.tl infile_list) @ included_files)
    in
    let _ = assert (included_files=[]) in
    let packed_list, unpacked_list = 
      List.fold_left 
        (fun (pl, upl) p -> 
          match p with
            | Packed p ->  p::pl, upl
            | Unpacked up -> pl, up::upl
        )
        ([], [])
        pack_list
    in
    let unpacked_merged_opt = (* All unpacked files are merged into one single package *)
      List.fold_left
        (fun acc pbody -> 
          match acc with
            | None -> Some pbody
            | Some pbody_acc -> 
              let add tbl x y =
                     (* Let's perform some clashes checks *)
                if Hashtbl.mem tbl x then
                  let ybis = Hashtbl.find tbl x in
		            print_string ("*** Error: "^(Ident.to_string x)^" is defined twice: \n\t" ^ 
                                   (Lxm.details y.src) ^ "\n\t" ^
                                   (Lxm.details ybis.src) ^ ".\n"); 
                  exit 2
                else
                  Hashtbl.add tbl x y
              in
              Hashtbl.iter (fun x y -> add pbody_acc.pk_const_table x y) pbody.pk_const_table;
              Hashtbl.iter (fun x y -> add pbody_acc.pk_type_table x y) pbody.pk_type_table;
              Hashtbl.iter (fun x y -> add pbody_acc.pk_node_table x y) pbody.pk_node_table;
              Some { pbody_acc with
                pk_def_list=pbody_acc.pk_def_list@pbody.pk_def_list;
              }
        )
        None
        unpacked_list
    in
    match unpacked_merged_opt with
      | None -> packed_list
      | Some unpacked_merged ->
        let name = 
          try Filename.chop_extension (Filename.basename first_file) 
          with _ -> 
		      print_string ("*** '"^first_file^"': bad file name.\n"); exit 1
        in
        let pi = AstV6.give_pack_this_name (Ident.pack_name_of_string name) unpacked_merged in
        let p = NSPack (Lxm.flagit pi (Lxm.dummy name)) in
        p::packed_list


let my_exit i =
  close_out !Global.oc;
  if Sys.file_exists !Global.outfile then Sys.remove !Global.outfile;
  exit i

let main = (
  (* Compile.init_appli () ; *)
  (* parse_args (); *)
  let args = MainArgs.parse Sys.argv in
  Verbose.exe ~level:3 (fun () ->
    Gc.set { (Gc.get ()) with Gc.verbose = 0x01 }
  );
  if !Global.run_unit_test then (
    UnifyType.unit_test ();
    exit 0
  );
  if (!Global.infiles = []) then (
    MainArgs.usage stderr args;
    exit 1
  );
  let main_node = 
    if !Global.main_node = "" then None else 
      Some (Ident.idref_of_string !Global.main_node)
  in
  if !Global.outfile <> "" then Global.oc := open_out !Global.outfile;
  try (
    let nsl = get_source_list !Global.infiles in
    let lic_prg = Compile.doit nsl main_node in

    if !Global.exec then
      (match main_node with
        | None -> (
          let first_file = List.hd !Global.infiles in
          let name = 
            try Filename.chop_extension (Filename.basename first_file) 
            with _ -> 
		        print_string ("*** '"^first_file^"': bad file name.\n"); exit 1
          in
          let nk = (Lic.node_key_of_idref (Ident.to_idref name)) in
          if LicPrg.node_exists lic_prg nk then (
            print_string ("WARNING: No main node is specified. I'll try with " ^ name ^"\n");
            flush stdout;
            let msk, zesoc = Lic2soc.f lic_prg nk in
            SocExec.f zesoc msk
          ) else (
            print_string ("Error: no node is specified, cannot exec.\n");
            flush stdout;
            exit 1
          )
        )
        | Some main_node -> 
          let msk, zesoc = Lic2soc.f lic_prg (Lic.node_key_of_idref main_node) in
          SocExec.f zesoc msk
      ) else (   
        LicPrg.to_file !Global.oc lic_prg main_node
       );

    Verbose.exe ~level:3 (fun () -> Gc.print_stat stdout);
  ) with
      Sys_error(s) ->
        prerr_string (s^"\n");
        my_exit 1
    | Global_error s ->
      print_global_error s ;
      my_exit 1
    | Parse_error ->
      print_compile_error (Lxm.last_made ()) "syntax error";
      my_exit 1
    | Unknown_var(lxm,id) -> 
      print_compile_error lxm ("unknown variable (" ^ (Ident.to_string id) ^")")
    | Unknown_constant(lxm,str) -> 
      print_compile_error lxm ("unknown constant (" ^ str ^")")
    | Compile_error(lxm,msg) -> 
      print_compile_error lxm msg;
      my_exit 1
    | L2lCheckLoops.Error(lxm,msg,lic_prg) -> 
      (* Sometime it helps to see the current state of the faulty program *)
      LicPrg.to_file !Global.oc lic_prg main_node;
      flush  !Global.oc;
      print_compile_error lxm msg;
      my_exit 1
    | Assert_failure (file, line, col)  -> 
      prerr_string (
        "\n*** oops: lus2lic internal error\n\tFile \""^ file ^
          "\", line " ^ (string_of_int line) ^ ", column " ^
          (string_of_int col) ^ "\n*** when compiling lustre program" ^
          (if List.length !Global.infiles > 1 then "s " else " ") ^
          (String.concat ", " !Global.infiles) ^ "\n"^
          "\n*** You migth want to sent a bug report to "^Version.maintainer ^"\n") ; 
      my_exit 2

(* | Compile_node_error(nkey,lxm,msg) -> ( *)
(* print_compile_node_error nkey lxm msg ; *)
(* exit 1 *)
(* ) *)
(* | Global_node_error(nkey,msg) -> ( *)
(* print_global_node_error nkey msg ; *)
(* exit 1 *)
(* ) *)

);
  close_out !Global.oc
