(* Time-stamp: <modified the 29/03/2013 (at 16:57) by Erwan Jahier> *)

open Soc
open SocExecValue

let dbg = Some(Verbose.get_flag "exec")

let (assign_expr : ctx -> var_expr -> var_expr -> ctx) =
  fun ctx ve_in ve_out -> (* ve_out := ve_in (in ctx) *)
    { ctx with s =
        let v = SocExecValue.get_value ctx ve_in in
        sadd_partial ctx.s ve_out ctx.cpath v 
    }

(* [array_index i v] returns the var_expr v[i] *)
let (array_index : int -> var -> var_expr) =
  fun i (vn,vt) -> 
    let vt_elt = 
      match vt with
        | Array(vt_elt,_) -> vt_elt
        | _ -> assert false
    in
    Index(Var(vn,vt),i,vt_elt)

let rec (soc_step : Soc.step_method -> Soc.tbl -> Soc.t -> SocExecValue.ctx
         -> SocExecValue.ctx) =
  fun step soc_tbl soc ctx ->
    let soc_name,_,_ = soc.key in
    match step.impl with
      | Predef -> (
        try SocExecEvalPredef.get soc.key ctx
        with Not_found -> (* Not a predef op *) print_string (
          "*** internal error in "^soc_name^". Is it defined in SocExecEvalPredef?\n");
          flush stdout; assert false
      )
      | Gaol(vl,gaol) -> List.fold_left (do_gao step.lxm soc_tbl) ctx gaol
      | Boolred(i,j,k) -> (
        let inputs, outputs = soc.profile in
        let b_array = (List.hd inputs) in
        let cpt = ref 0 in
        for i = 0 to k-1 do
          let a_i = array_index i b_array in
          let v = SocExecValue.get_value ctx a_i in
          if v = B true then incr cpt;
        done;
        let res = B (!cpt >= i && !cpt <= j) in
        let res_var = fst (List.hd outputs) in
        let s = sadd ctx.s (res_var::ctx.cpath) res in
        { ctx with s = s }
      )
      | Iterator(iter, node_sk, n) -> 
        let node_soc = SocUtils.find step.lxm node_sk soc_tbl in
        let node_step = match node_soc.step with [step] -> step | _ ->  assert false in
        let iter_inputs,iter_outputs = soc.profile in
        let rctx = ref ctx in
        let (proc_name,_,_) = node_soc.key in 
        let inst_name =
          match soc.instances with
            | []  -> Array.make n proc_name
            | _   -> Array.of_list (List.map fst soc.instances)
        in
        for i = 0 to n-1 do
          rctx := { !rctx with cpath = inst_name.(i)::ctx.cpath };
          let vel_in, vel_out =
            match iter with
              | "map" -> (List.map (array_index i) iter_inputs,
                          List.map (array_index i) iter_outputs)
              | "fold" | "red" | "fill" | "fillred"  -> 
                let a_in = Var (List.hd iter_inputs) in
                ( a_in::(List.map (array_index i) (List.tl iter_inputs)),
                  a_in::(List.map (array_index i) (List.tl iter_outputs)))
              | _ -> assert false (* should not occur *)
          in
          rctx := do_step inst_name.(i) node_step !rctx soc_tbl node_soc vel_in vel_out;
        done;
        if iter <> "map" then (
          let a_in = Var (List.hd iter_inputs) in
          let a_out = Var (List.hd iter_outputs) in
          rctx := assign_expr !rctx a_in a_out);  (* a_out=a_n *)
        !rctx;


and (do_gao :  Lxm.t -> Soc.tbl -> SocExecValue.ctx -> gao -> SocExecValue.ctx) =
  fun lxm soc_tbl ctx gao ->
    match gao with
      | Case(id, id_gao_l) -> (
        let id_val = get_enum id ctx in
        let gaol = try List.assoc id_val id_gao_l  with Not_found -> assert false in
        List.fold_left (do_gao lxm soc_tbl) ctx gaol
      )
      | Call(vel_out, Assign, vel_in) -> List.fold_left2 assign_expr ctx vel_in vel_out
      | Call(vel_out, Procedure sk, vel_in) -> (
        let (proc_name,_,_) = sk in
        let ctx = { ctx with cpath = proc_name::ctx.cpath } in
        let soc = SocUtils.find lxm sk soc_tbl in
        let step = match soc.step with [step] -> step | _ ->  assert false in
        do_step proc_name step ctx soc_tbl soc vel_in vel_out 
      )
      | Call(vel_out, Method((inst_name,sk),step_name), vel_in) -> (
        let ctx = { ctx with cpath = inst_name::ctx.cpath } in
        let soc = SocUtils.find lxm sk soc_tbl in
        let step = try List.find (fun sm -> sm.name = step_name) soc.step
          with Not_found -> assert false
        in
        do_step inst_name step ctx soc_tbl soc vel_in vel_out
      )
and (do_step : Ident.t -> step_method -> SocExecValue.ctx -> Soc.tbl -> Soc.t -> 
     var_expr list -> var_expr list -> SocExecValue.ctx) =
  fun name step ctx soc_tbl soc vel_in vel_out -> 
    let soc_in_vars, soc_out_vars = soc.profile in
    let step_in_vars = filter_params soc soc_in_vars step.idx_ins in 
    let step_out_vars = filter_params soc soc_out_vars step.idx_outs in
    let new_s = substitute_args_and_params vel_in step_in_vars ctx in
    let ctx = soc_step step soc_tbl soc { ctx with s=new_s } in
    let s_out = substitute_params_and_args step_out_vars vel_out ctx in
    { s = s_out ; cpath = List.tl ctx.cpath }


(* get the step params from its soc params *)
and (filter_params : Soc.t -> Soc.var list -> int list -> Soc.var list) =
  fun soc el il ->
    let local_nth i l = 
      if i < 0 then 
        (* to handle the particular case of arrow. XXX just a crutch to make it works *)
        let mem_name = SocPredef.get_mem_name soc.Soc.key Soc.Bool in
        (mem_name, Soc.Bool)
      else
        try List.nth l i 
        with _ ->
          print_string (
            "\n*** Cannot get the " ^ (string_of_int (i+1)) ^ 
              "th element of a list of size " ^ (string_of_int (List.length l))^"\n");
          flush stdout;
          assert false 
    in
    let res  = List.map (fun i -> local_nth i el) il  in
    res

(* expand struct and arrays when communicating with the outside world (a good idea?) *)
let rec (expand_profile:Soc.var list -> Soc.var list) =
  fun vl -> 
    let res = List.flatten (List.map expand_var vl) in
    (* fix point. now useless ? *)
    if List.length res = List.length vl then res else
      expand_profile res
and expand_var var = match var with
  | (vn,(Bool| Int | Real)) -> [var]
  | (vn,Enum(n,l)) ->  [vn,Int] 
  | (vn,Array(vt,i)) ->
    let res = ref [] in
    for k=i-1 downto 0 do
      res := (vn^"_"^(string_of_int k),vt) :: !res;
    done;
    (expand_profile !res)
  | (vn,Struct(name,fl)) -> expand_profile (List.map (fun (fn,t) -> vn^"_"^fn,t ) fl)
  | (vn,Extern id) -> assert false (* finish me! *)
  | (vn,Alpha _) -> assert false (* should not occur *)

let (int_to_enum : SocExecValue.t -> Soc.ident list -> SocExecValue.t) =
  fun v el -> 
    match v with
      | I i -> (try E (List.nth el i,i) with _ -> 
        failwith ("Enum out of the range [0,"^(string_of_int (List.length el))^"]"))
      | _ -> assert false (* should not occur *)

let rec (expand_subst: Rif_base.subst -> Rif_base.subst list) =
  fun s ->
    let rec aux acc (n,v) =
      match v with
        | U | I _ | F _ | B _  -> (n,v)::acc
        | E(_e,i) -> (n,I i)::acc
        | S fl -> 
          let f (fn,fv) = n^"_"^fn, fv in
          let fl = List.map f fl in
          List.fold_left aux acc fl
        | A a -> 
          let res = ref acc in
          for i=0 to (Array.length a)-1 do
            let n_i = n^"_"^(string_of_int i) in
            res := aux !res (n_i, a.(i));
          done;
          !res
    in
    aux [] s

(* A local shortcut to ease the profile def *)
type sl = Rif_base.subst list

(* Reconstruct the flattenned data *)
let (unexpand : sl -> Soc.var list -> sl) =
  fun sl vl -> 
    let rec (aux : sl -> sl -> Soc.var list -> sl * sl)=
      fun sl_done sl_todo vl -> 
        (* Returns the (accumulated) result and the unused subst
           (which should be empty at the top-level call) *)
        match sl_todo, vl with
          | _,[] -> sl_done, sl_todo
          | s::sl, (_, (Bool| Int | Real))::vl -> aux (s::sl_done) sl vl
          | (id,v)::sl, (_,Enum(n,el))::vl -> 
            let s =  (id, int_to_enum v el) in
            aux (s::sl_done) sl vl 
          | _, (vn, Array(vt,i))::vl -> (
            let sl_todo_ref = ref sl_todo in
            let sl_done_ref = ref [] in
            let a_fake_value = I 42 in
            let res = Array.make i a_fake_value in
            for k=0 to i-1 do
              let (vk_l:Soc.var list) = [("fake_name",vt)] in
              let (sl_done_v, sl_todo_v) = aux !sl_done_ref !sl_todo_ref vk_l in
              sl_todo_ref:=sl_todo_v;
              sl_done_ref:=sl_done_v;
              Array.set res k (snd (List.hd !sl_done_ref));
            done; 
            let sl_done = (vn, A res)::sl_done in
            aux sl_done !sl_todo_ref vl
          )
          | _, (vn,Struct(sn,fl))::vl -> 
            let sl_todo, fl = List.fold_left aux_field (sl_todo,[]) fl in
            let sl_done = (vn, S fl)::sl_done in
            aux sl_done sl_todo vl

          | _, (vn,Extern id)::_ -> assert false (* finish me! *)
          | _, (vn,Alpha _  )::_ -> assert false (* should not occur *)
          | [],_::_ -> assert false  (* should not occur *)

    and (aux_field : sl * (ident * SocExecValue.t) list -> ident * var_type 
         -> sl * (ident * SocExecValue.t) list ) =
      fun (sl_todo, fl) (fn, t) ->
        let new_sl_done, sl_todo = aux [] sl_todo [fn,t] in
        let (_,v) = List.hd new_sl_done in
        sl_todo, (fn,v)::fl

    in
    let res, remaining = aux [] sl vl in
    assert (remaining=[]);
    res

let (read_soc_input : Soc.t -> out_channel -> substs -> substs) =
  fun soc oc ctx_s -> 
    let profile = expand_profile (fst soc.profile) in
    let vntl = List.map (fun (x,t) -> x,SocUtils.string_of_type_ref t) profile in
    let s:Rif_base.subst list = Rif_base.read stdin (Some oc) vntl in
    let s = unexpand s (fst soc.profile) in
    List.fold_left (fun acc (id,v) -> sadd acc [id] v) ctx_s s

let rec (loop_step : Soc.tbl -> Soc.t -> SocExecValue.ctx -> int -> out_channel -> unit) =
  fun soc_tbl soc ctx step_nb oc ->
    Rif_base.write oc ("\n#step " ^ (string_of_int step_nb)^"\n");
    let ctx = { ctx with s = read_soc_input soc oc ctx.s } in
    let step = match soc.step with [step] -> step | _ ->  assert false in
    let ctx = soc_step step soc_tbl soc ctx in
(*     dump_substs  ctx.s; *)
    let profile = expand_profile (snd soc.profile) in
    let vntl = List.map (fun (x,t) -> x,SocUtils.string_of_type_ref t) profile in
    let s = SocExecValue.filter_top_subst ctx.s in
    let s = List.flatten(List.map expand_subst s) in
    Rif_base.write oc " #outs ";
    Rif_base.write_outputs oc vntl s;
    Rif_base.flush oc;
    Verbose.exe ~flag:dbg (fun () -> dump_substs ctx.s; flush stdout);
    loop_step soc_tbl soc ctx (step_nb+1) oc

let (f : Soc.tbl -> Soc.key -> unit) =
  fun soc_tbl sk ->
    let soc = try SocMap.find sk soc_tbl with Not_found -> assert false in
    let ctx = SocExecValue.create_ctx soc_tbl soc in
    let vntl_of_profile = List.map (fun (x,t) -> x,SocUtils.string_of_type_ref t) in
    let vntl_i = vntl_of_profile (expand_profile (fst soc.profile)) in
    let vntl_o = vntl_of_profile (expand_profile (snd soc.profile)) in
    let oc = 
      if  !Global.outfile  = "" then stdout else
      let rif_file =
        try (Filename.chop_extension !Global.outfile) ^ ".rif" 
        with _  ->  !Global.outfile ^ ".rif"
      in
      open_out rif_file 
    in
    Verbose.dump_entete oc;
    Rif_base.write_interface oc vntl_i vntl_o None None;
    Rif_base.flush oc;
    try loop_step soc_tbl soc ctx 1 oc
    with Rif_base.Bye -> 
      close_out oc
