(* Time-stamp: <modified the 29/03/2013 (at 16:55) by Erwan Jahier> *)

open Soc

type t = I of int | F of float | B of bool | E of Soc.ident * int
         | A of t array | S of (Soc.ident * t) list | U

(* Meant to represent paths in the call tree. Actually it both
   represent path and variable with a path, depending on the
   context *)
type path = ident list

let (path_to_string: ident list -> string) =
  fun l -> 
    String.concat "->" (List.rev l)

type subst = (path * t)

(* soc environnement holding memory values.

   Each soc have a list of memories, which is actually a list of soc
   instances (in the licc terminology). 

   Hence we need a kind of mapping from instance list (the path in the
   call tree) to values.
*)
(* type substs = subst list *)
type substs =
  | Node of (ident * substs) list
  | Leaf of t

type ctx = { 
  cpath: path; 
  s:substs;
}

(**************************************************************************)

let rec (get_top_var_type : Soc.var_expr -> Soc.var_type) =
  fun ve -> 
    match ve with
      | Var(_,vt)  -> vt
      | Index(ve,_,_)  
      | Field(ve, _, _)  -> get_top_var_type ve
      | Const(id,_) -> assert false

type access = Idx of int | Fld of ident

let rec (get_access : Soc.var_expr -> access list) =
  fun ve -> 
    match ve with
      | Var(id,_)  -> []
      | Index(ve,i,_)  -> (Idx i)::(get_access ve)
      | Field(ve, n,_)  -> (Fld n)::(get_access ve)
      | Const(id,_) -> assert false

(* Replace access(pre_v) by v in pre_v *)
let rec (update_val : t -> t -> access list -> t) =
  fun pre_v v access -> 
    match pre_v,access with
      | _,[] -> v
      | A a, (Idx i)::access -> 
        let a_i = update_val a.(i) v access in
        a.(i) <- a_i;
        A a
      | S(fl), (Fld fn)::access -> 
        S (List.map 
             (fun (fn2,v2) -> if fn=fn2 then fn,update_val v2 v access else (fn2,v2)) 
             fl)
      | _,_ -> assert false (* finish me (field struct) *)

let (update_leaf : var_expr -> t -> t -> substs) =
  fun ve v pre_v -> 
    let access = get_access ve in
    let new_v = update_val pre_v v access in
    Leaf(new_v)

(* The same as update in the case where no previous value exists *)
let rec (create_val : Soc.var_type -> t ->  access list -> t) =
  fun vt v access ->
    match vt,access with
      | _,[] -> v
      | Array(vt,size), (Idx i)::access -> 
        let a = Array.make size U in
        let a_i = create_val vt v access in
        a.(i) <- a_i;
        A a
      | Struct(sn,fl), (Fld fn)::access -> 
        S(List.map (fun (fn2,vt2) -> if fn=fn2 then fn,create_val vt2 v access else fn2,U) fl)
      | _,_ -> assert false

let (create_leaf : var_expr -> t -> substs) =
  fun ve v -> 
    let access = get_access ve in
    let top_vt = get_top_var_type ve in
    let new_v = create_val top_vt v access in
    Leaf(new_v)

let rec (get_top_id : Soc.var_expr -> ident) =
  function
  | Var(id,_) | Const(id,_) -> id
  | Field(ve, _, _) | Index(ve,_,_) -> get_top_id ve



(* exported *)
let (sadd_partial : substs  -> var_expr -> path -> t -> substs) =
  fun ct ve x v ->
    let top_id = get_top_id ve in
    let x = top_id::x in
    let rec aux ct (x,v) =
    match ct,x with
      | Leaf(pre_v),[] -> update_leaf ve v pre_v
      | Node(l),n::t -> (
        try 
          let s = aux (List.assoc n l) (t,v) in
          Node((n,s)::(List.remove_assoc n l))
        with Not_found -> 
          if t = [] then Node((n, create_leaf ve v)::l)
          else
          let new_substs = aux (Node []) (t,v) in
          Node((n,new_substs)::l)
      )
      | _,[] -> assert false
      | Leaf(_),_ -> assert false
    in
    aux ct (List.rev x,v)

(* let rec (sadd : substs -> subst -> substs) =  *)
(*   fun ct (x,v) -> *)
(*     (x,v)::(List.remove_assoc x ct) *)

(* exported *)
let (sadd : substs -> path -> t -> substs) = 
  fun ct x v ->
    let rec aux ct (x,v) =
    match ct,x with
      | Leaf(_),[] -> Leaf(v)
      | Node(l),n::t -> (
        try 
          let s = aux (List.assoc n l) (t,v) in
          Node((n,s)::(List.remove_assoc n l))
        with Not_found -> 
          if t = [] then Node((n,Leaf v)::l)
          else
          let new_substs = aux (Node []) (t,v) in
          Node((n,new_substs)::l)
      )
      | _,[] -> assert false
      | Leaf(_),_ -> assert false
    in
    aux ct (List.rev x,v)


(**************************************************************************)


(* let filter_top_subst s = *)
(*   List.fold_left *)
(*     (fun acc (idl,v) ->  *)
(*         match idl with  *)
(*           | [id] -> (id,v)::acc  *)
(*           | _ -> acc  *)
(*     ) *)
(*     [] *)
(*     s *)

let (filter_top_subst : substs -> (Soc.ident * t) list) =  
  fun s -> 
    let rec aux acc (id, s) =
      match s with
        | Leaf(v) -> (id,v)::acc
        | _ -> acc
    in
    match s with 
        Node(l) -> List.fold_left aux [] l
      | _ -> assert false

let rec (to_string  : t -> string) =
  function 
    | I i  -> string_of_int i
    | F f  -> string_of_float f
    | B true -> "t"
    | B false -> "f"
    | E (e,_) -> e
    | S fl -> String.concat " " (List.map (fun (fn,fv) -> to_string fv) fl)
    | A a -> 
      let str = ref "" in
      let f i a = str := !str ^ " " ^ (to_string a) in
      Array.iteri f a;
      !str
    | U -> "not initialised"

let (string_of_subst_list : (path * t) list -> string) =
  fun  s -> 
    let values = List.map (fun (var,value) -> (path_to_string var)^"="^(to_string value)) s in
     ((String.concat "\n\t" values) ^ "\n")

let (dump_subst_list : (path * t) list -> unit) =
  fun s -> 
    print_string (string_of_subst_list s);
    flush stdout


(* let (substs_to_list: substs -> (path * t) list) = *)
(*   fun s -> s *)

let (substs_to_list: substs -> (path * t) list) =
  fun s -> 
    let rec aux acc s p =
      match s with 
        | Node(l) -> List.fold_left (fun acc (id,s) -> aux acc s (id::p)) acc l
        | Leaf(v) -> (p,v)::acc
    in
    aux [] s []

let (string_of_substs : substs -> string) =
  fun s -> 
    string_of_subst_list (substs_to_list s)

let (dump_substs : substs -> unit) =
  fun s -> 
    dump_subst_list (substs_to_list s)


(* XXX use a Rif reader *)
let rec (read_enum : ident list -> ident) = 
  fun idl -> 
    print_string ("Enter " ^ (String.concat "," idl)); flush stdout; 
    let str = read_line () in
    if List.mem str idl then str else (print_string "Bad enum; "; read_enum idl)

let rec pos_in_list i x l =
  match l with
    | e::l -> if e=x then i else pos_in_list (i+1) x l
    | [] -> assert false (* should not occur *)


let (read_value : var -> t) =
  fun (_,t) -> 
    match t with
      | Bool -> print_string "Enter a bool (t/f):";flush stdout; B(read_line () = "t") 
      | Int  -> print_string "Enter an int:"; flush stdout; I(read_int())
      | Real -> print_string "Enter a float:"; flush stdout; F(read_float()) 
      | Enum(_,idl)  -> 
        let e = read_enum(idl) in
        let i = pos_in_list 0 e idl in
          E(e,i)
      | _ -> assert false (* finish me! *)

(* let (get_val : ident -> ctx -> t) = *)
(*   fun id ctx ->  *)
(*     try List.assoc (id::ctx.cpath) ctx.s *)
(*     with Not_found ->  *)
(*       let msg = (path_to_string (id::ctx.cpath)) ^ " unbound in "  *)
(*         ^ (string_of_substs ctx.s)  *)
(*       in *)
(*       print_string msg; flush stdout; *)
(*       assert false *)


let (get_val : ident -> ctx -> t) =
  fun id ctx -> 
    let rec find ct p =
      match ct,p with
        | Node(l),n::t -> find (List.assoc n l) t
        | Leaf(v),[] -> v
        | _,[] -> raise Not_found
        | Leaf(_),_ -> assert false
    in
    try find ctx.s (List.rev (id::ctx.cpath))
    with Not_found -> 
      let msg = (path_to_string (id::ctx.cpath)) ^ " unbound in \n" 
        ^ (string_of_substs ctx.s) 
      in
      print_string msg; flush stdout;
      assert false

let (get_enum : ident -> ctx -> ident) =
fun id ctx -> 
  match get_val id ctx with
    | E(e,_) -> e
    | _ -> assert false (* should not fail *) 

let rec (get_value : ctx -> var_expr -> t) =
  fun ctx v -> 
    match v with
      | Var(id,_) -> get_val id ctx
      | Const("true",Bool) ->  B true
      | Const("false",Bool) -> B false
      | Const(id_in,Int)  -> I (int_of_string id_in)
      | Const(id_in,Real) -> F (float_of_string id_in)
      | Const(id,Enum(_,idl)) -> E(id, pos_in_list 0 id idl)
      | Const(id,_) -> assert false  (* finish me! *)
      | Field(ve,fn,t) -> 
        let s = get_value ctx ve in
        (match s with
          | S fl -> (
            try List.assoc fn fl
            with Not_found -> assert false (* should not occur *)
          )
          | _ -> assert false (* should not occur *)
        )

      | Index(ve,i,vt) -> 
        let a = get_value ctx ve in
        match a with
          | A a -> a.(i)
          | _ -> assert false (* should not occur *)


(* type substs = subst list *)
(* let (sadd : substs -> subst -> substs) =  *)
(*   fun ss (x,v) ->  *)
(*     let ss = List.remove_assoc x ss in *)
(*     (x,v)::ss *)

(* let (string_of_substs :substs -> string) =  *)
(*   fun s ->  *)
(*     let str_l =  *)
(*       List.map (fun (var,value) -> (path_to_string var) ^ "=" ^ (to_string value)) s  *)
(*     in *)
(*     "{\n\t"^ (String.concat "\n\t" str_l) ^ " }" *)
(*  *)
(*  *)
(* let (get_val : ident -> ctx -> t) = *)
(*   fun id ctx ->  *)
(*     try List.assoc (id::ctx.cpath) ctx.s *)
(*     with Not_found ->  *)
(*       let msg = (path_to_string (id::ctx.cpath)) ^ " unbound in " ^ (string_of_substs ctx.s) in *)
(*       print_string msg; flush stdout; *)
(*       assert false *)



(* exported *)
let (substitute_args_and_params : var_expr list -> var list -> ctx -> substs) =
  fun args params ctx -> 
    assert (List.length args = List.length params);
    let arg_ctx = { ctx with cpath = List.tl ctx.cpath } in
    let s = List.fold_left2
      (fun acc arg (pn,_) -> sadd acc (pn::ctx.cpath) (get_value arg_ctx arg))
      ctx.s args params 
    in
    s

let (substitute_params_and_args : var list -> var_expr list -> ctx -> substs) = 
  fun params args ctx -> 
    assert (List.length args = List.length params);
    let s = List.fold_left2
      (fun acc arg (pn,_) -> sadd_partial acc arg (List.tl ctx.cpath) (get_val pn ctx) )
      ctx.s args params
    in
    s



let empty_ctx: ctx = {
  cpath = [];
  s     = Node [];
}

(* exported *)
let rec (create_ctx : Soc.tbl -> Soc.t -> ctx) =
  fun soc_tbl soc -> 
    let rec (init_soc: Soc.t -> ident list -> substs -> substs) =
      fun soc cpath mem ->
        let mem =
          match soc.have_mem with
            | Some(vt, Some(value)) ->
              let name = (SocPredef.get_mem_name soc.key vt)::cpath in
              let value = get_value empty_ctx value in
              sadd mem name value
            | Some(vt, None) ->
              let name = (SocPredef.get_mem_name soc.key vt)::cpath in
              let value = U in
              sadd mem name value
            | None -> mem
        in
        List.fold_left (init_instances cpath) mem soc.instances

    and (init_instances: ident list -> substs -> Soc.instance -> substs) =
      fun cpath mem (iname, sk) ->
        let soc = SocUtils.find_no_exc sk soc_tbl in
        init_soc soc (iname::cpath) mem
    in
    {
      s = init_soc soc [] (Node []);
      cpath = [];
    }

